@@ -4,6 +4,7 @@ from typing import Type
 import numpy as np
 import pandas as pd
 import pytest
+from sklearn.base import is_classifier
 from sklearn.multioutput import ClassifierChain, MultiOutputClassifier
 
 import sklearndf.classification as classification
@@ -43,15 +44,6 @@ CLASSIFIER_INIT_PARAMETERS = {
 }
 
 
-@pytest.mark.parametrize(argnames="sklearndf_cls", argvalues=CLASSIFIERS_TO_TEST)
-def test_wrapped_constructor(sklearndf_cls: Type[ClassifierDF]) -> None:
-    """ Test standard constructor of wrapped sklearn classifiers """
-    # noinspection PyArgumentList
-    _: ClassifierDF = sklearndf_cls(
-        **CLASSIFIER_INIT_PARAMETERS.get(sklearndf_cls.__name__, {})
-    )
-
-
 @pytest.mark.parametrize(argnames="sklearndf_cls", argvalues=CLASSIFIERS_TO_TEST)
 def test_wrapped_fit_predict(
     sklearndf_cls: Type[ClassifierDF],
@@ -66,6 +58,8 @@ def test_wrapped_fit_predict(
         **CLASSIFIER_INIT_PARAMETERS.get(sklearndf_cls.__name__, {})
     )
 
+    assert is_classifier(classifier)
+
     is_chain = isinstance(classifier.native_estimator, ClassifierChain)
 
     is_multi_output = isinstance(classifier.native_estimator, MultiOutputClassifier)
